// This file is part of the FlexVer™ reference implementation
//
// © 2017 - 2018 Raptor Engineering, LLC
// All Rights Reserved
//
// WARNING
// This material may be covered by patent(s) and/or patent(s) pending
//
// Any use of this file without an active FlexVer™ license from
// Raptor Engineering, LLC is strictly prohibited.  A single
// exception to this is made for the sole purpose of third-party
// security auditing of FlexVer™ and associated technologies.
//
// For the security audit exception to apply, the primary purpose
// of the use must be to audit the security of the FlexVer™ technology,
// and Raptor Engineering, LLC must be notified in writing of
// any security flaw(s) or potential security flaw(s) found during
// the audit no less than four (4) weeks in advance of public
// disclosure of said flaw(s).  Furthermore, the number of compiled
// FlexVer™ binaries must be kept to the minimum necessary to conduct
// the security audit, and all compiled FlexVer™ binaries must be
// destroyed once the audit has been completed.  These restrictions
// do not apply to organizations or individuals possessing an active
// FlexVer™ license from Raptor Engineering, LLC.
//
// For sales and licensing information, please visit
// https://www.raptorengineering.com
//
// PUBLIC VERSION FOR GENERAL REDISTRIBUTION
// Licensed for use with Raptor Data Solutions, LLC's IntegriCloud™ service

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <unistd.h>
#include <arpa/inet.h>

#include <uECC.h>
#include <sha2.h>

#include "libflexver.h"

#define FLEXVER_IDLE_OP_TIMEOUT	30
#define FLEXVER_LOCK_OP_TIMEOUT	30

#define CONSOLE_NORMAL		"\x1B[0m"

#define CONSOLE_FONT_BLACK	"\x1B[30m"
#define CONSOLE_FONT_RED	"\x1B[31m"
#define CONSOLE_FONT_GREEN	"\x1B[32m"
#define CONSOLE_FONT_YELLOW	"\x1B[33m"
#define CONSOLE_FONT_BLUE	"\x1B[34m"
#define CONSOLE_FONT_MAGENTA	"\x1B[35m"
#define CONSOLE_FONT_CYAN	"\x1B[36m"
#define CONSOLE_FONT_WHITE	"\x1B[37m"
#define CONSOLE_FONT_NONE	""

#define CONSOLE_BG_BLACK	"\x1B[40m"
#define CONSOLE_BG_RED		"\x1B[41m"
#define CONSOLE_BG_GREEN	"\x1B[42m"
#define CONSOLE_BG_YELLOW	"\x1B[43m"
#define CONSOLE_BG_BLUE		"\x1B[44m"
#define CONSOLE_BG_MAGENTA	"\x1B[45m"
#define CONSOLE_BG_CYAN		"\x1B[46m"
#define CONSOLE_BG_WHITE	"\x1B[47m"
#define CONSOLE_BG_NONE		""

#define CONSOLE_STEADY		""
#define CONSOLE_BLINK		"\x1B[5m"

flexver_log_level flexver_debug_level = FLEXVER_LOGLEVEL_NONE;

void flexver_log(flexver_log_level log_level, const char* format, ...) {
	va_list args;
	va_start(args, format);
	if (flexver_debug_level >= log_level) {
		vprintf(format, args);
	}
	va_end(args);
}

struct sort_pair {
	uint8_t value;
	uint8_t frequency;
} sort_pair_t;

void set_flexver_loglevel(flexver_log_level log_level) {
	flexver_debug_level = log_level;
}

uint8_t get_secure_state_from_flexver_status_field(uint32_t status, uint8_t signature_valid, uint8_t assert_security_violation) {
	uint32_t secure;

	if (signature_valid) {
		secure = 1;
	}
	else {
		secure = 0;
	}

	if (assert_security_violation) {
		secure = 0;
	}

	if (status & FLEXVER_STATUS_INITIALIZED) {
		//
	}
	else {
		secure = 0;
	}
	if (status & FLEXVER_STATUS_GENERATING_PUBLIC_KEY) {
		if (secure) secure = 2;
	}
	if (status & FLEXVER_STATUS_GENERATING_SIGNATURE) {
		if (secure) secure = 2;
	}
	if (status & FLEXVER_STATUS_GENERATING_HASH) {
		if (secure) secure = 2;
	}
	if (status & FLEXVER_STATUS_CHECKING_FIRMWARE) {
		if (secure) secure = 2;
	}
	if (status & FLEXVER_STATUS_SECURE) {
		//
	}
	else {
		secure = 0;
	}
	if (status & FLEXVER_STATUS_HOST_FW_UPDATE) {
		secure = 0;
	}
	if (status & FLEXVER_STATUS_HOST_ONLINE) {
		//
	}
	else {
		if (secure) secure = 2;
	}
	if (status & FLEXVER_STATUS_HOST_SECURE) {
		//
	}
	else {
		secure = 0;
	}
	if (status & FLEXVER_STATUS_HASH_OBSOLETE) {
		secure = 0;
	}
	if (status & FLEXVER_STATUS_HOST_RESET_ACTIVE) {
		if (secure) secure = 2;
	}
	if (status & FLEXVER_STATUS_HOST_RESET_MISSED) {
		secure = 0;
	}
	if (status & FLEXVER_STATUS_HOST_POWER_ACTIVE) {
		//
	}
	else {
		if (secure) secure = 2;
	}
	if (status & FLEXVER_STATUS_KEY_STORAGE_LOCKED) {
		//
	}
	else {
		if (secure) secure = 3;
	}
	if (status & FLEXVER_STATUS_SEC_HASH_STORAGE_LOCKED) {
		//
	}
	else {
		if (secure) secure = 3;
	}
	if (status & FLEXVER_STATUS_TRANSFERRING_FIRMWARE) {
		if (secure) secure = 2;
	}

	return secure;
}

// NOTE
// The returned string(s) must be freed with free() after use!
char* decode_flexver_status_field(uint32_t status, uint8_t signature_valid, uint8_t assert_security_violation, uint8_t format_dual_line, char** second_line) {
	uint32_t buffer_size = (128 * 16) + 1;	// Up to 64 characters for each status flag (64 characters plus control codes), plus null terminator
	char* buffer = malloc(buffer_size);
	char* ret_ptr = buffer;
	uint32_t length = 0;
	uint32_t secure = get_secure_state_from_flexver_status_field(status, signature_valid, assert_security_violation);

	if (!buffer) {
		return NULL;
	}

	memset(buffer, 0, buffer_size);

	if (format_dual_line) {
		*second_line = malloc(buffer_size);
		if (!(*second_line)) {
			free(buffer);
			return NULL;
		}
		memset(*second_line, 0, buffer_size);
	}

	if (secure == 0) {
		sprintf(buffer + strlen(buffer), "%s%s%s SECURITY VIOLATION %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_RED, CONSOLE_BLINK, CONSOLE_NORMAL);
	}
	else if (secure == 1) {
		sprintf(buffer + strlen(buffer), "%s%s%s SYSTEM SECURE %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_GREEN, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	else if (secure == 2) {
		sprintf(buffer + strlen(buffer), "%s%s%s SYSTEM SECURITY STATE UNKNOWN %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_YELLOW, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	else if (secure == 3) {
		sprintf(buffer + strlen(buffer), "%s%s%s CONNECTION SECURITY UNKNOWN %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_YELLOW, CONSOLE_STEADY, CONSOLE_NORMAL);
	}

	if (status & FLEXVER_STATUS_INITIALIZED) {
		sprintf(buffer + strlen(buffer), "%s%s%s FLEXVER ACTIVE %s|", CONSOLE_FONT_GREEN, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_GENERATING_PUBLIC_KEY) {
		sprintf(buffer + strlen(buffer), "%s%s%s GENERATING PUBLIC KEY %s|", CONSOLE_FONT_WHITE, CONSOLE_BG_BLUE, CONSOLE_BLINK, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_GENERATING_SIGNATURE) {
		sprintf(buffer + strlen(buffer), "%s%s%s GENERATING SIGNATURE %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_CYAN, CONSOLE_BLINK, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_GENERATING_HASH) {
		sprintf(buffer + strlen(buffer), "%s%s%s CALCULATING HASH %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_MAGENTA, CONSOLE_BLINK, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_CHECKING_FIRMWARE) {
		sprintf(buffer + strlen(buffer), "%s%s%s VERIFYING HOST FIRMWARE %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_MAGENTA, CONSOLE_BLINK, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_SECURE) {
		sprintf(buffer + strlen(buffer), "%s%s%s FLEXVER SECURE %s|", CONSOLE_FONT_GREEN, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	else {
		sprintf(buffer + strlen(buffer), "%s%s%s FLEXVER SECURITY VIOLATION %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_RED, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_HOST_FW_UPDATE) {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST FIRMWARE UPDATE %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_YELLOW, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_HOST_ONLINE) {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST ONLINE %s|", CONSOLE_FONT_GREEN, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	else {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST OFFLINE %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_RED, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_HOST_SECURE) {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST BOOT SECURED %s|", CONSOLE_FONT_GREEN, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	else {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST INSECURE %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_RED, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_HASH_OBSOLETE) {
		sprintf(buffer + strlen(buffer), "%s%s%s HASH OBSOLETE %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_YELLOW, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	else {
		sprintf(buffer + strlen(buffer), "%s%s%s HASH VALID %s|", CONSOLE_FONT_GREEN, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (signature_valid) {
		sprintf(buffer + strlen(buffer), "%s%s%s SIGNATURE VALID %s|", CONSOLE_FONT_GREEN, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	else {
		sprintf(buffer + strlen(buffer), "%s%s%s SIGNATURE INVALID %s|", CONSOLE_FONT_RED, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}

	if (format_dual_line) {
		length = strlen(buffer);
		if (length > 0) {
			buffer[length-1] = 0;
		}

		buffer = *second_line;
	}

	if (status & FLEXVER_STATUS_KEY_STORAGE_LOCKED) {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST KEYSTORE LOCKED %s|", CONSOLE_FONT_GREEN, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	else {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST KEYSTORE UNLOCKED %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_YELLOW, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_SEC_HASH_STORAGE_LOCKED) {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST HASHSTORE LOCKED %s|", CONSOLE_FONT_GREEN, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	else {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST HASHSTORE UNLOCKED %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_YELLOW, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_HOST_RESET_ACTIVE) {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST RESET ACTIVE %s|", CONSOLE_FONT_NONE, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_HOST_RESET_MISSED) {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST RESET FAULT %s|", CONSOLE_FONT_RED, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_HOST_POWER_ACTIVE) {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST POWER ACTIVE %s|", CONSOLE_FONT_GREEN, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_HOST_STANDBY_POWER_ACT) {
		sprintf(buffer + strlen(buffer), "%s%s%s HOST STANDBY POWER ACTIVE %s|", CONSOLE_FONT_GREEN, CONSOLE_BG_NONE, CONSOLE_STEADY, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_TRANSFERRING_FIRMWARE) {
		sprintf(buffer + strlen(buffer), "%s%s%s TRANSFERRING HOST FIRMWARE %s|", CONSOLE_FONT_BLACK, CONSOLE_BG_MAGENTA, CONSOLE_BLINK, CONSOLE_NORMAL);
	}
	if (status & FLEXVER_STATUS_OPERATION_QUEUED) {
		sprintf(buffer + strlen(buffer), "%s%s%s REQUEST QUEUED %s|", CONSOLE_FONT_NONE, CONSOLE_BG_NONE, CONSOLE_BLINK, CONSOLE_NORMAL);
	}

	length = strlen(buffer);
	if (length > 0) {
		buffer[length-1] = 0;
	}

	return ret_ptr;
}

int extract_flexver_ssh_public_key_components(uint8_t* flexver_ssh_public_key, uint8_t* exponent, uint8_t* modulus, uint32_t* exponent_length, uint32_t* modulus_length) {
	int i;
	uint32_t length_le;
	memcpy(modulus, flexver_ssh_public_key, 512);
	memcpy(exponent, flexver_ssh_public_key + 512, 16);
	*modulus_length = ntohl(*((uint32_t*)(flexver_ssh_public_key + (512 + 16))));
	*exponent_length = ntohl(*((uint32_t*)(flexver_ssh_public_key + (512 + 16 + 4))));
}

int verify_flexver_core_integrity_measurements(uint32_t flexver_status, uint16_t flexver_internal_boot_counter, uint8_t* flexver_public_key, uint8_t* flexver_signature, uint8_t* flexver_primary_firmware_hash, uint8_t* flexver_secondary_firmware_hash, uint8_t* flexver_signature_nonce, uint8_t* flexver_ssh_public_key, uint8_t* signature_valid, uint32_t* flexver_status_field, uint16_t* flexver_internal_boot_count) {
	int i;

	// Assemble full message for verification
	uint8_t flexver_signed_message[670];
	uint8_t flexver_message_hash[32];
	uint8_t flexver_signature_valid;
	SHA256_CTX flexver_message_hash_sha256_context;
	memcpy(flexver_signed_message, flexver_ssh_public_key, 536);
	memcpy(flexver_signed_message + 536, flexver_primary_firmware_hash, 32);
	memcpy(flexver_signed_message + 568, flexver_secondary_firmware_hash, 32);
	memcpy(flexver_signed_message + 600, flexver_signature_nonce, 64);
	flexver_signed_message[664] = (flexver_status >> 24) & 0xff;
	flexver_signed_message[665] = (flexver_status >> 16) & 0xff;
	flexver_signed_message[666] = (flexver_status >> 8) & 0xff;
	flexver_signed_message[667] = flexver_status & 0xff;
	flexver_signed_message[668] = (flexver_internal_boot_counter >> 8) & 0xff;
	flexver_signed_message[669] = flexver_internal_boot_counter & 0xff;

	// Override signature generation status flags to match state of system
	// during signature computation
	flexver_signed_message[664] |= (FLEXVER_STATUS_OPERATION_QUEUED >> 24);
	flexver_signed_message[667] |= FLEXVER_STATUS_GENERATING_SIGNATURE;

	// Compute SHA256 hash of full message
	SHA256_Init(&flexver_message_hash_sha256_context);
	SHA256_Update(&flexver_message_hash_sha256_context, flexver_signed_message, 670);
	SHA256_Final(flexver_message_hash, &flexver_message_hash_sha256_context);

	flexver_log(FLEXVER_LOGLEVEL_INFO, "[INFO] Verifying FlexVer™ signature...");
	flexver_log(FLEXVER_LOGLEVEL_INFO, "\n[INFO] Assembled message:\n[INFO] [%02x", flexver_signed_message[0]);
	for (i = 1; i < 670; i++) {
		flexver_log(FLEXVER_LOGLEVEL_INFO, " %02x", flexver_signed_message[i]);
	}
	flexver_log(FLEXVER_LOGLEVEL_INFO, "]\n");
	flexver_log(FLEXVER_LOGLEVEL_INFO, "\n[INFO] Message hash:\n[INFO] [%02x", flexver_message_hash[0]);
	for (i = 1; i < 32; i++) {
		flexver_log(FLEXVER_LOGLEVEL_INFO, " %02x", flexver_message_hash[i]);
	}
	flexver_log(FLEXVER_LOGLEVEL_INFO, "]\n");
	uECC_Curve ecdsa_curve = uECC_secp256r1();
	flexver_signature_valid = uECC_verify(flexver_public_key, flexver_message_hash, 32, flexver_signature, ecdsa_curve);
	if (flexver_signature_valid) {
		flexver_log(FLEXVER_LOGLEVEL_INFO, "[INFO] Signature valid!\n");
	}
	else {
		flexver_log(FLEXVER_LOGLEVEL_INFO, "[INFO] Signature INVALID\n");
	}

	if (signature_valid) {
		*signature_valid = flexver_signature_valid;
	}
	if (flexver_status_field) {
		*flexver_status_field = flexver_status;
	}
	if (flexver_internal_boot_count) {
		*flexver_internal_boot_count = flexver_internal_boot_counter;
	}
	return 0;
}
